@@ -1,6 +1,6 @@
 /*
   JavaScript utilities for PmWiki
-  (c) 2009-2025 Petko Yotov www.pmwiki.org/petko
+  (c) 2009-2026 Petko Yotov www.pmwiki.org/petko
   based on PmWiki addons DeObMail, AutoTOC and Ape
   licensed GNU GPLv2 or any more recent version released by the FSF.
 
@@ -8,14 +8,14 @@
   a Public Domain event listener by github.com/tofsjonas
 */
 
-
-
 (async function(__script__){
   
-  const { Now, echo, err, aE, dE, dCE, qsa, dqs, dqsa, dce, tap,
-    abs, floor, floatval, min, zpad, minmax, mkDrag,
-    PHSC, getConf, jP, posy, adjany, adjbb, adjbe, adjab, adjae,
-    gcs, css, getLS, setLS, delLS, ready } = PmLib;
+  const { echo, err, Now, qs, qsa, dqs, dqsa, PHSC, dce,
+    css, min, max, minmax, floor, floatval, intval,
+    abs, zpad, getLS, setLS, jP, getConf, aE, dE, tap,
+    adjbb, adjbe, adjab, adjae, posy, mkDrag,
+    gcs, fe4j, fe4b, fe4t, fe4, fe4p, ready } = PmLib;
+    
   var Config = getConf(__script__);
   
   if(Config.rediquiet) {
@@ -23,19 +23,17 @@
     if(url != location.href) 
       history.replaceState(null, null, url);
   }
-  
-  
-  
+
   var wikitext;
 
   function PmXMail() {
-    var els = document.querySelectorAll('span._pmXmail');
+    var els = dqsa('span._pmXmail');
     var LinkFmt = '<a href="%u" class="mail">%t</a>';
 
     for(var i=0; i<els.length; i++) {
-      var x = els[i].querySelector('span._t');
+      var x = qs(els[i], 'span._t');
       var txt = cb_mail(x.innerHTML);
-      var y = els[i].querySelector('span._m');
+      var y = qs(els[i], 'span._m');
       var url = cb_mail(y.innerHTML.replace(/^ *-&gt; */, ''));
 
       if(!url) url = 'mailto:'+txt.replace(/^mailto:/, '');
@@ -58,11 +56,11 @@
   
   function any_id(h) {
     if(h.id) {return h.id;} // %id=anchor%
-    var a = h.querySelector('a[id]'); // inline [[#anchor]]
+    var a = qs(h, 'a[id]'); // inline [[#anchor]]
     if(a && a.id) {return a.id;}
     var prev = h.previousElementSibling;
     if(prev) { // [[#anchor]] before !!heading
-      var a = prev.querySelectorAll('a[id]');
+      var a = qsa(prev, 'a[id]');
       if(a.length) {
         last = a[a.length-1];
         if(last.id && ! last.nextElementSibling) {
@@ -78,6 +76,7 @@
     return false;
   }
 
+  
   function autotoc() {
     if(dqs('.noPmTOC')) { return; } // (:notoc:) in page
     var dtoc = Config.pmtoc;
@@ -89,6 +88,7 @@
         if(specs[i].match(/^[1AI]$/i)) numheadspec[i] = ''+specs[i];
         else numheadspec[i] = '';
       }
+      if(!dtoc.MinNumberedHeadings) dtoc.MinNumberedHeadings = 2;
     }
 
     var query = [];
@@ -96,7 +96,7 @@
       query.push('h'+i);
     }
     if(dtoc.EnableQMarkup) query.push('p.question');
-    var pageheadings = wikitext.querySelectorAll(query.join(','));
+    var pageheadings = qsa(wikitext, query.join(','));
     if(!pageheadings.length) { return; }
 
     var toc_headings = [ ];
@@ -110,7 +110,8 @@
 
     var tocdiv = dqs('.PmTOCdiv');
     var shouldmaketoc = ( tocdiv || (toc_headings.length >= dtoc.MinNumber && dtoc.MinNumber != -1)) ? 1:0;
-    if(!dtoc.NumberedHeadings && !shouldmaketoc) return;
+    var shouldnumheadings = (dtoc.NumberedHeadings && toc_headings.length >= dtoc.MinNumberedHeadings) ? 1:0;
+    if(!shouldnumheadings && !shouldmaketoc) return;
 
     for(var i=0; i<toc_headings.length; i++) {
       var h = toc_headings[i];
@@ -120,10 +121,17 @@
       var id = any_id(h);
       hcache.push([h, level, id]);
     }
+    
+    var blink = false;
+    if(dtoc.EnableBacklinks) {
+      let bltip = dtoc.EnableBacklinks === 1 ? dtoc.contents:dtoc.EnableBacklinks;
+      blink = dce('a', { href:"#_toc", className:"back-arrow", title:bltip });
+    }
 
     prevlevel = 0;
     var html = '';
     var toclinks = [];
+    let frag = dce('template');
     for(var i=0; i<hcache.length; i++) {
       var hc = hcache[i];
       var actual_level = hc[1] - minlevel;
@@ -133,33 +141,38 @@
         hc[2] = 'toc-'+currnb.id.replace(/\.+$/g, '');
         hc[0].id = hc[2];
       }
-      var txt = hc[0].textContent.trim().replace(/</g, '&lt;');
-      var sectionedit = hc[0].querySelector('.sectionedit');
-      if(sectionedit) {
-        var selength = sectionedit.textContent.length;
-        txt = txt.slice(0, -selength);
-      }
-      var nhtml = currnb.num.replace(/^((?:\w+\.)+)(\w+\.)$/, 
-        '<span class="pmparnum">$1</span>$2');
-      nhtml = `<span class="pmnum pmnum${actual_level}">${nhtml} </span>`;
       
-      if(dtoc.NumberedHeadings && currnb.num.length) {
+      if(shouldmaketoc) {
+        frag = frag.cloneNode();
+        frag.innerHTML = hc[0].innerHTML;
+        
+        let se = qsa(frag.content, '.sectionedit, .rmTOC');
+        for(let se1 of se) se1.remove();
+        let ke = qsa(frag.content, '.kpTOC');
+        for(let ki=0; ki<ke.length; ki++) {
+          let ker = dce('span', {textContent:`[kT${ki}kT]`});
+          adjbb(ke[ki], ker);
+          ke[ki].remove();
+        }
+        let txt = PHSC(frag.content.textContent.trim());
+        if(ke.length) txt = txt.replace(/\[kT(\d+)kT\]/g, function(a, a1){
+          return ke[intval(a1)].outerHTML;
+        });
+        
+        var toclink = dce('a', {
+          className:'pmtoc-indent' + actual_level,
+          href: '#' + hc[2],
+          innerHTML: txt
+        });
+        toclinks.push(toclink);
+        if(blink) adjbe(hc[0], blink.cloneNode());
+      }
+      if(shouldnumheadings) {
+        let alevel1 = actual_level+1;
+        nhtml = `<span class="pmnumwrap pmnumwrap${alevel1}">${currnb.num} </span>`;
         adjab(hc[0], nhtml);
+        if(shouldmaketoc) adjab(toclink, nhtml);
       }
-
-      if(! shouldmaketoc) { continue; }
-      
-      var a = dce('a', {
-        className:'pmtoc-indent' + actual_level,
-        href: '#' + hc[2],
-        textContent: txt
-      });
-      if(dtoc.NumberedHeadings && currnb.num.length) {
-        adjab(a, nhtml);
-      }
-      toclinks.push(a);
-      if(dtoc.EnableBacklinks) 
-        adjbe(hc[0], '<a class="back-arrow" href="#_toc"></a>');
     }
 
     if(! shouldmaketoc) return;
@@ -168,7 +181,7 @@
       id: '_toc',
       className: 'PmTOCdiv frame'
     });
-    if(! getLS('closeTOC')) details.open = true;
+    if(! getLS('closeTOC', dtoc.Closed)) details.open = true;
  
     var sum = dce('summary', {textContent: dtoc.contents});
     details.appendChild(sum);
@@ -201,7 +214,7 @@
 
     var hh = location.hash;
     if(hh.length>1) {
-      var cc = document.getElementById(hh.substring(1));
+      var cc = dqs(hh);
       if(cc) cc.scrollIntoView();
     }
   }
@@ -240,13 +253,14 @@
     var qq = '';
     var hh = '';
     for (var j=0; j<=n; j++) {
-      var curr = numhead[j];
+      let j1 = j+1;
+      let curr = numhead[j];
       hh += curr+".";
-      var currspec = numheadspec[j];
+      let currspec = numheadspec[j];
       if(!currspec.match(/[1ai]/i)) continue;
       if(currspec.match(/a/i)) { curr = numhead_alpha(curr, currspec); }
       else if(currspec.match(/i/i)) { curr = numhead_roman(curr, currspec); }
-      qq+=curr+".";
+      qq+= `<span class="pmnum${j1}">${curr}.</span>`;
       
     }
     return {id:hh.replace(/\.$/, ''), num:qq};
@@ -256,20 +270,25 @@
     if(! floatval(Config.sortable)) return;
     var tables = dqsa('table.sortable,table.sortable-footer');
     for(var i=0; i<tables.length; i++) {
-      var rows = tables[i].querySelectorAll('tr');
+      if(qs(tables[i], '[rowspan],[colspan]')) {
+        tables[i].classList.replace('sortable', 'unsortable');
+        tables[i].classList.replace('sortable-footer', 'unsortable-footer');
+        continue;
+      }
+      var rows = qsa(tables[i], 'tr');
       // non-pmwiki-core table, already ready
-      if(! tables[i].querySelector('thead')) {
+      if(! qs(tables[i], 'thead')) {
         var thead = dce('thead');
-        tables[i].insertBefore(thead, tables[i].firstChild);
+        adjab(tables[i], thead);
         thead.appendChild(rows[0]);
       }
-      var tbody = tables[i].querySelector('tbody');
+      var tbody = qs(tables[i], 'tbody');
       if(! tbody) {
         tbody = dce('tbody');
         tables[i].appendChild(tbody);
         for(var r=1; r<rows.length; r++) tbody.appendChild(rows[r]);
       }
-      if(tables[i].className.match(/sortable-footer/)) {
+      if(tables[i].classList.contains('sortable-footer')) {
         tables[i].classList.add('sortable');
         var tfoot = dce('tfoot');
         tables[i].appendChild(tfoot);
@@ -280,15 +299,15 @@
     libsortable();
   }
   function mkdatasort(rows) {
-    var hcells = rows[0].querySelectorAll('th,td');
+    var hcells = qsa(rows[0], 'th,td');
     var specialsort = [], span;
     for(var i=0; i<hcells.length; i++) {
-      sortspan = hcells[i].querySelector('.sort-number,.sort-number-us,.sort-date,.sort-time');
+      sortspan = qs(hcells[i], '.sort-number,.sort-number-us,.sort-date,.sort-time');
       specialsort[i] = sortspan? sortspan.className : false;
     }
     if(! specialsort.length) return;
     for(var i=1; i<rows.length; i++) {
-      var cells = rows[i].querySelectorAll('td,th');
+      var cells = qsa(rows[i], 'td,th');
       var k = 0;
       for(var j=0; j<cells.length && j<specialsort.length; j++) {
         if(! specialsort[j]) continue;
@@ -297,8 +316,8 @@
         else if(specialsort[j] == 'sort-number') {ds = t.replace(/[^-,\d]+/g, '').replace(/,/g, '.'); }
         else if(specialsort[j] == 'sort-date') {ds = new Date(t).getTime(); }
         else if(specialsort[j] == 'sort-time') {
-          var time = cells[j].querySelector('time[datetime]');
-          if(time) ds = time.getAttribute('datetime');
+          var time = qs(cells[j],'time[datetime]');
+          if(time) ds = time.datetime;
         }
         if(ds) cells[j].dataset.sort = ds;
       }
@@ -309,7 +328,7 @@
     // updated to keep any event handlers in the table
     function getValue(obj) {
       obj = obj.cells[column_index];
-      return obj.getAttribute('data-sort') || obj.innerText;
+      return obj.dataset.sort || obj.innerText;
     }
     function reclassify(element, cname) {
       element.classList.remove('dir-u', 'dir-d');
@@ -320,17 +339,15 @@
       if(e.target.closest('a')) return; // links
       var element = e.target.closest('th');
       if(! element) return;
-      var table = element.offsetParent;
+      var table = element.closest('table');
       if(!table.classList.contains('sortable')) return;
-                              
-      var cells = element.parentNode.cells;
-      for (var i = 0; i < cells.length; i++) {
-        if(cells[i] === element) {
-          column_index = i;
-        } else {
-          reclassify(cells[i], '');
-        }
-      }
+
+      var prev = qsa(table, 'th.dir-d, th.dir-u');
+      for (let cell of prev) if(cell !== element) reclassify(cell, '');
+
+      var cells = [...element.parentNode.cells];
+      column_index = cells.indexOf(element);
+
       var cname = 'dir-d', reverse = false;
       if(element.classList.contains(cname)) {
         cname = 'dir-u';
@@ -338,9 +355,8 @@
       }
       reclassify(element, cname);
       var tbody = table.tBodies[0];
-      var rows = [];
-      for(var r=0; r<tbody.rows.length; r++) rows.push(tbody.rows[r]);
-      
+      var rows = [...tbody.rows];
+
       rows.sort(function(x, y) {
         var a = getValue(reverse? y:x),
             b = getValue(reverse? x:y);
@@ -348,9 +364,8 @@
             d = a - b;
         return isNaN(d) ? c : d;
       });
-      for (i = 0; i < rows.length; i++) {
-        tbody.appendChild(rows[i]);
-      }
+      for (let row of rows) tbody.appendChild(row);
+
     });
   }
 
@@ -381,12 +396,12 @@
     
     for(var i=0; i<x.length; i++) {
       if(x[i].className.match(/(^| )(pm|pmwiki)( |$)/)) { continue;} // core highlighter
-      var pre = Array.from(x[i].querySelectorAll('pre,code'));
+      var pre = [...qsa(x[i],'pre,code')];
       var n = x[i].nextElementSibling;
       if(n && n.tagName == 'PRE') pre.push(n);
       for(var j=0; j<pre.length; j++) {
         pre[j].className += ' ' + x[i].className;
-        var varlinks = pre[j].querySelectorAll('a.varlink');
+        var varlinks = qsa(pre[j], 'a.varlink');
         var vararray = {};
         for(var v=0; v<varlinks.length; v++) {
           vararray[varlinks[v].textContent] = varlinks[v].href;
@@ -395,7 +410,7 @@
         if(pre[j].children) pre[j].textContent = pre[j].textContent;
 
         hljs.highlightElement(pre[j]);
-        var hlvars = pre[j].querySelectorAll('span.hljs-variable');
+        var hlvars = qsa(pre[j], 'span.hljs-variable');
         for(var v=0; v<hlvars.length; v++) {
           var hlvar = hlvars[v].textContent;
           if(vararray.hasOwnProperty(hlvar)) {
@@ -404,7 +419,7 @@
         }
       }
     }
-    var varlinks = dqsa('a.varlink code.varlink');
+    var varlinks = dqsa('code.varlink');
     for(i=0; i<varlinks.length; i++) varlinks[i].classList.add('hljs-variable');
   }
   
@@ -448,7 +463,7 @@
     var h72 = Now.getTime()/1000-days*24*3600;
     
     for(var i=0; i<times.length; i++) {
-      var itemdate = new Date(times[i].dateTime);      
+      var itemdate = new Date(times[i].dateTime);
       var stamp = floor(itemdate.getTime()/1000);
       
       var li = times[i].closest('li');
@@ -458,7 +473,7 @@
         times[i].title = x[1] ? x[1]: itemdate.textContent;
         continue;
       }
-      var link = li.querySelector('a');
+      var link = qs(li, 'a');
       if(link.className.match(/createlinktext|wikilink|selflink/)) {
         var diff = link.href + '?action=diff#diff' + stamp;
         if(stamp >= h72) {
@@ -482,11 +497,11 @@
       var x = fmtLocalTime(stamp);
       
       link.innerHTML = x[0];
-      link.setAttribute('title', x[1] ? x[1]: link.textContent);
+      link.title = x[1] ? x[1]: link.textContent;
       
       var par = link.closest('li');
       if(!par) continue;
-      par.insertBefore(link, par.firstChild);
+      adjab(par, link);
       adjae(link, "&nbsp;&nbsp;");
       if(previous && stamp>previous) par.classList.add('rcnew');
     }
@@ -599,158 +614,165 @@
     }
   }
   
-  var Dropzone = false;
-  var UploadQueue = [];
   function init_dropzone(){
     if(!Config.updrop) return;
-    var form = dqs('form[action$="action=postupload"], form[action$="action=edit"]');
-    if(!form) return;
-    Dropzone = dce('div', {className: "frame pmdropzone"});
-    
-    var label = dce('span', {textContent: Config.updrop.label});
-    adjbe(label, '&nbsp;');
-    Dropzone.appendChild(label);
-    if(Config.updrop.areq) {
-      var areq = form.querySelector('input[name="author"]')
-      Dropzone.pmareq = areq;
+    var forms = dqsa('form[action$="action=postupload"], form[action$="action=edit"]');
+    for (let f of forms) {
+      var div = dce('div', {className: "frame pmdropzone" });
+      var label = dce('span', {textContent: Config.updrop.label+' '});
+      adjab(div, label);
+      adjab(f, div);
     }
- 
-    adjab(form, Dropzone);
-    aE(Dropzone, 'dragenter', dragenter);
-    aE(Dropzone, 'dragover', dragenter);
+    var zones = dqsa('.pmdropzone,div[data-pmdropzone]');
+    for(let z of zones) mkdropzone(z);
+  }
+  
+  function mkdropzone(Dropzone) {
+    var UploadQueue = [];
+    
+    aE(Dropzone, 'dragenter,dragover', dragenter);
     aE(Dropzone, 'dragleave', dragleave);
     aE(Dropzone, 'drop', dragdrop);
     tap(Dropzone, clickzone);
-  }
-  
-  async function postUpload(file) {
-    var url = Config.updrop.action;
-    var f = new FormData;
-    f.append('n', Config.fullname);
-    f.append('action', 'postupload');
-    f.append('uploadfile', file);
-    f.append('pmdrop', 1);
-    f.append(Config.updrop.token[0], Config.updrop.token[1]);
-    if(Dropzone.pmareq) 
-      f.append('author', Dropzone.pmareq.value);
-    try {
-      var response = await fetch(url, { method: 'POST', body: f });
-      if(!response.ok) {
-        var msg = `HTTP error! Status: ${response.status}`;
-        throw new Error(msg);
-        return { error: 1, msg: msg};
-      }
-      var txt =  await response.text();
-      var json = jP(txt, {error: 1, msg: txt });
-      return json;
+    
+    const conf = jP(Dropzone.dataset.pmdropzone, {});
+    
+
+    function findauthor(){
+      if(conf.author) return conf.author;
+      let authors = dqsa('input[name="author"]');
+      for(let a of authors) if(a.value) return a.value;
+      return '';
     }
-    catch(er) {
-      err(er);
-      return {error: 1, msg: 'Error posting form data: ' + er.message };
+    function dragenter(e) {
+      e.preventDefault();
+      Dropzone.classList.add('over');
     }
-  }
-  
-  async function processUploads() {
-    while(true) {
-      var a = dqs('.pmdropzone a.queued');
-      if(!a) return;
-      a.className = 'uploading';
-      var result = await postUpload(a.pmfile);
-      if(result.error) {
-        a.className = 'error';
-        a.title = result.msg;
-      }
-      else {
-        a.className = 'success';
-        a.href = result.href;
-        a.textContent = result.uprname;
-        a.title = result.msg;
-        delete a.pmfile;
-      }
+    
+    function dragleave(e) {
+      e.preventDefault();
+      Dropzone.classList.remove('over');
     }
-  }
-  
-  function clickzone(e){
-    var a = e.target.closest('a');
-    if(!a) return;
-    e.preventDefault();
-    e.stopPropagation();
-    if(a.className.match(/uploading|queued|deleting/)) return;
-    else if(a.className == 'success' && typeof insMarkup == 'function') {
-      var pn = e.ctrlKey? Config.fullname + '/': '';
-      var text = "Attach:"+pn+a.textContent;
-      if(text.match(/\s/)) text = '[['+text+']]';
-      insMarkup(function(x){
-        if(x.length) return text;
-        return {
-          mtext: text,
-          unselect: true
-        };
-      });
-      return;
-    }
-    else if(a.className == 'error' || typeof insMarkup == 'undefined') {
-      a.classList.add('deleting');
-      setTimeout(function(){a.remove();}, 700);
-      return;
-    }
-    // this shouldn't happen
-  }
-  
-  function dragenter(e) {
-    e.preventDefault();
-    Dropzone.classList.add('over');
-  }
-  
-  function dragleave(e) {
-    e.preventDefault();    
-    Dropzone.classList.remove('over');
-  }
-  
-  async function dragdrop(e) {
-    e.preventDefault();   
-    Dropzone.classList.remove('over');
-    if(Dropzone.pmareq) {
-      var v = Dropzone.pmareq.value;
-      if(!v.length) {
-        appendUpload({size:-500,name:Config.updrop.areq});
-        Dropzone.pmareq.focus();
+    
+    async function dragdrop(e) {
+      e.preventDefault();
+      Dropzone.classList.remove('over');
+      let author = findauthor();
+      if(Config.updrop.areq && !author) {
+        appendUpload({size: -500, name: Config.updrop.areq});
         return;
       }
+      var files = event.dataTransfer.files;
+      for (var i = 0; i < files.length; i++) appendUpload(files[i]);
+      var pending = dqs('.pmdropzone a.uploading');
+      if(!pending) await processUploads();
     }
-    var files = event.dataTransfer.files;
-    for (var i = 0; i < files.length; i++) appendUpload(files[i]);
-    var pending = dqs('.pmdropzone a.uploading');
-    if(!pending) await processUploads();
+    
+    function appendUpload(file) {
+      if(file.size==0) { return; }
+      var sizes = Config.updrop.sizes;
+      var ext = '';
+      var m = file.name.match(/\.([^\.]+)$/);
+      if(m) ext  = m[1].toLowerCase();
+      
+      var a = dce('a', {href: '#', textContent: file.name});
+      
+      if(file.size==-500) {
+        a.className = 'error';
+      }
+      else if(typeof sizes[ext] == 'undefined') {
+        a.className = 'error';
+        a.title = Config.updrop.badtype.replace(/\#upext/, ext);
+      }
+      else if(file.size > sizes[ext]) {
+        a.className = 'error';
+        a.title = Config.updrop.toobig
+        .replace(/\#upmax/, sizes[ext]).replace(/\#upext/, ext);
+      }
+      else {
+        a.className = 'queued';
+        a.pmfile = file;
+      }
+      Dropzone.appendChild(a);
+    }
+    
+    async function postUpload(file) {
+      var url = conf.action? conf.action : Config.updrop.action;
+      var f = new FormData;
+      f.append('n', conf.n? conf.n : Config.fullname);
+      f.append('action', 'postupload');
+      f.append('uploadfile', file);
+      f.append('pmdrop', 1);
+      f.append(Config.updrop.token[0], Config.updrop.token[1]);
+      f.append('author', findauthor());
+      try {
+        var response = await fetch(url, { method: 'POST', body: f });
+        if(!response.ok) {
+          var msg = `HTTP error! Status: ${response.status}`;
+          throw new Error(msg);
+          return { error: 1, msg: msg};
+        }
+        var txt =  await response.text();
+        var json = jP(txt, {error: 1, msg: txt });
+        return json;
+      }
+      catch(er) {
+        err(er);
+        return {error: 1, msg: 'Error posting form data: ' + er.message };
+      }
+    }
+    
+    async function processUploads() {
+      while(true) {
+        var a = dqs('.pmdropzone a.queued');
+        if(!a) return;
+        a.className = 'uploading';
+        var result = await postUpload(a.pmfile);
+        if(result.error) {
+          a.className = 'error';
+          a.title = result.msg;
+        }
+        else {
+          a.className = 'success';
+          a.href = result.href;
+          a.textContent = result.uprname;
+          a.title = result.msg;
+          delete a.pmfile;
+        }
+      }
+    }
+    
+    function clickzone(e) {
+      var a = e.target.closest('a');
+      if(!a) return;
+      e.preventDefault();
+      e.stopPropagation();
+      if(a.className.match(/uploading|queued|deleting/)) return;
+      else if(a.className == 'success' && typeof insMarkup == 'function') {
+        var fullname = conf.n? conf.n: Config.fullname;
+        var pn = conf.n || e.ctrlKey? fullname + '/': '';
+        var text = "Attach:"+pn+a.textContent;
+        if(text.match(/\s/)) text = '[['+text+']]';
+        insMarkup(function(x){
+          if(x.length) return text;
+          return {
+            mtext: text,
+            unselect: true
+          };
+        });
+        return;
+      }
+      else if(a.className == 'error' || typeof insMarkup == 'undefined') {
+        a.classList.add('deleting');
+        setTimeout(function(){a.remove();}, 700);
+        return;
+      }
+      // this shouldn't happen
+    }
+    
   }
   
-  function appendUpload(file) {
-    if(file.size==0) { return; }
-    var sizes = Config.updrop.sizes;
-    var ext = '';
-    var m = file.name.match(/\.([^\.]+)$/);
-    if(m) ext  = m[1].toLowerCase();
-
-    var a = dce('a', {href: '#', textContent: file.name});
-    
-    if(file.size==-500) {
-      a.className = 'error';
-    }
-    else if(typeof sizes[ext] == 'undefined') {
-      a.className = 'error';
-      a.title = Config.updrop.badtype.replace(/\#upext/, ext);
-    }
-    else if(file.size > sizes[ext]) {
-      a.className = 'error';
-      a.title = Config.updrop.toobig
-        .replace(/\#upmax/, sizes[ext]).replace(/\#upext/, ext);
-    }
-    else {
-      a.className = 'queued';
-      a.pmfile = file;
-    }
-    Dropzone.appendChild(a);
-  }
+  
   
   function init_draggable() {
     var els = dqsa('.draggable');
@@ -772,10 +794,14 @@
       try{
         var target = dqs(a[3]);
         if(!target) continue;
+        if(el.contains(target)) {
+          err("Shifted contains Target!", {el, target});
+          continue;
+        }
         target.insertAdjacentElement(a[2], el);
         if(a[1]) {
           var inner = [...el.childNodes];
-          for(var node of inner) el.parentNode.insertBefore(node, el);
+          for(var node of inner) adjbb(el, node);
           el.remove();
           return;
         }
@@ -827,6 +853,56 @@
     aE(y, 'wheel', wheelinputone, {passive: false});
   }
   
+  function intercheckboxes(){
+    let shiftKey = 0;
+    let lastChecked = null;
+    aE('body', 'keydown', function(e){
+      if(e.key == 'Shift') shiftKey = 1;
+    });
+    aE('body', 'keyup', function(e){
+      if(e.key == 'Shift') shiftKey = 0;
+    });
+    aE('body', 'change', function(e){
+      const box = e.target;
+      if(box.type !== 'checkbox') return;
+      if(shiftKey && lastChecked) {
+        const all = [...dqsa('input[type="checkbox"]')];
+        const x = all.indexOf(lastChecked);
+        if(x>=0) {
+          const y = all.indexOf(box);
+          const a = min(x, y);
+          const b = max(x, y);
+          const checked = box.checked;
+          for (let i = a; i <= b; i++) {
+            all[i].checked = checked;
+          }
+        }
+      }
+      lastChecked = box;
+    });
+  }
+  
+  function confirmUnload() {
+    let forms = dqsa('form[data-confirmunload]');
+    if(!forms.length) return;
+    aE(forms, 'change', function(e){ this.__changed = 1; });
+    aE(forms, 'submit', function(e){ this.__submitting = 1;
+      setTimeout(function(){delete this.__submitting; }, 500);
+    });
+    
+    aE(window, 'beforeunload', function(e){
+      for(let f of forms) {
+        let changed = (f.__changed || f.dataset.confirmunload == '2');
+        if(changed && !f.__submitting) {
+          e.preventDefault();
+          e.returnValue = '';
+          return;
+        }
+      }
+    });
+  }
+  
+  
   ready(function(){
     wikitext = document.getElementById('wikitext');
     var fn = [
@@ -835,7 +911,8 @@
       highlight_pre, copy_pre, 
       confirmForms, labelForms, wheelinputinit,
       init_dropzone, init_draggable,
-      makesortable
+      makesortable, intercheckboxes,
+      confirmUnload
     ];
     fn.forEach(function(a){a();});
     setTimeout(datashift, 800);
